import numpy as np
import pytest
from unittest.mock import patch, MagicMock
from datetime import datetime
import geopandas as gpd
import pandas as pd
import pytz
from os import path

from metloom.pointdata import (
    CDECPointData,
    PointDataCollection,
    PointData,
)
from metloom.variables import CdecStationVariables


def side_effect_error(*args):
    raise ValueError("Testing error")


class TestPointData(object):
    @pytest.fixture(scope="class")
    def data_dir(self):
        this_dir = path.dirname(__file__)
        return path.join(this_dir, "data")

    @pytest.fixture(scope="class")
    def shape_obj(self, data_dir):
        fp = path.join(data_dir, "testing.shp")
        return gpd.read_file(fp)

    @staticmethod
    def expected_response(dates, vals, var, unit, station, points):
        obj = []
        for dt, v in zip(dates, vals):
            obj.append(
                {
                    "datetime": pd.Timestamp(dt, tz="UTC"),
                    "measurementDate": pd.Timestamp(dt, tz="UTC"),
                    var: v,
                    f"{var}_units": unit,
                    "site": station.id,
                }
            )
        df = gpd.GeoDataFrame.from_dict(
            obj,
            geometry=[points] * len(dates),
        )
        # needed to reorder the columns for the pd testing compare
        df = df.filter(
            ["datetime", "geometry", "site", "measurementDate", var, f"{var}_units"]
        )
        df.set_index(keys=["datetime", "site"], inplace=True)
        return df


class TestCDECStation(TestPointData):
    @staticmethod
    def cdec_daily_response():
        return [
            {
                "stationId": "TNY",
                "durCode": "D",
                "SENSOR_NUM": 3,
                "sensorType": "SNOW WC",
                "date": "2021-5-16 00:00",
                "obsDate": "2021-5-16 00:00",
                "value": -0.11,
                "dataFlag": " ",
                "units": "INCHES",
            },
            {
                "stationId": "TNY",
                "durCode": "D",
                "SENSOR_NUM": 3,
                "sensorType": "SNOW WC",
                "date": "2021-5-17 00:00",
                "obsDate": "2021-5-17 00:00",
                "value": -0.10,
                "dataFlag": " ",
                "units": "INCHES",
            },
            {
                "stationId": "TNY",
                "durCode": "D",
                "SENSOR_NUM": 3,
                "sensorType": "SNOW WC",
                "date": "2021-5-18 00:00",
                "obsDate": "2021-5-18 00:00",
                "value": -0.10,
                "dataFlag": " ",
                "units": "INCHES",
            },
        ]

    @pytest.fixture(scope="function")
    def tny_station(self):
        return CDECPointData("TNY", "Tenaya Lake")

    @pytest.fixture(scope="class")
    def tny_daily_expected(self):
        points = gpd.points_from_xy([-119.0], [42.0], z=[1000.0])
        df = gpd.GeoDataFrame.from_dict(
            [
                {
                    "datetime": pd.Timestamp("2021-05-16 07:00:00+0000", tz="UTC"),
                    "measurementDate": pd.Timestamp(
                        "2021-05-16 07:00:00+0000", tz="UTC"
                    ),
                    "ACCUMULATED PRECIPITATION": -0.11,
                    "ACCUMULATED PRECIPITATION_units": "INCHES",
                    "site": "TNY",
                },
                {
                    "datetime": pd.Timestamp("2021-05-17 07:00:00+0000", tz="UTC"),
                    "measurementDate": pd.Timestamp(
                        "2021-05-17 07:00:00+0000", tz="UTC"
                    ),
                    "ACCUMULATED PRECIPITATION": -0.10,
                    "ACCUMULATED PRECIPITATION_units": "INCHES",
                    "site": "TNY",
                },
                {
                    "datetime": pd.Timestamp("2021-05-18 07:00:00+0000", tz="UTC"),
                    "measurementDate": pd.Timestamp(
                        "2021-05-18 07:00:00+0000", tz="UTC"
                    ),
                    "ACCUMULATED PRECIPITATION": -0.10,
                    "ACCUMULATED PRECIPITATION_units": "INCHES",
                    "site": "TNY",
                },
            ],
            geometry=[points[0]] * 3,
        )
        # needed to reorder the columns for the pd testing compare
        df = df.filter(
            [
                "datetime",
                "geometry",
                "site",
                "measurementDate",
                "ACCUMULATED PRECIPITATION",
                "ACCUMULATED PRECIPITATION_units",
            ]
        )
        df.set_index(keys=["datetime", "site"], inplace=True)
        return df

    @classmethod
    def tny_side_effect(cls, url, **kwargs):
        mock = MagicMock()
        if kwargs["params"].get("dur_code") == "D":
            mock.json.return_value = cls.cdec_daily_response()
        elif kwargs["params"].get("dur_code") == "H":
            raise NotImplementedError()
        elif "getStationInfo" in url:
            mock.json.return_value = {
                "STATION": [
                    {
                        "SENS_LONG_NAME": "SNOW, WATER CONTENT",
                        "ELEVATION": 1000.0,
                        "LATITUDE": 42.0,
                        "LONGITUDE": -119.0,
                    }
                ]
            }
        else:
            raise ValueError("unknown scenario")
        return mock

    @pytest.fixture(scope="class")
    def station_search_response(self):
        return [
            pd.DataFrame.from_records(
                [
                    (
                        "GIN",
                        "GIN FLAT",
                        "MERCED R",
                        "MARIPOSA",
                        -119.773,
                        37.767,
                        7050,
                        "CA Dept of Water Resources/DFM-Hydro-SMN",
                        np.nan,
                    ),
                    (
                        "DAN",
                        "DANA MEADOWS",
                        "TUOLUMNE R",
                        "TUOLUMNE",
                        -119.257,
                        37.897,
                        9800,
                        "CA Dept of Water Resources/DFM-Hydro-SMN",
                        np.nan,
                    ),
                    (
                        "TNY",
                        "TENAYA LAKE",
                        "MERCED R",
                        "MARIPOSA",
                        -119.448,
                        37.838,
                        8150,
                        "CA Dept of Water Resources/DFM-Hydro-SMN",
                        np.nan,
                    ),
                    (
                        "GFL",
                        "GIN FLAT (COURSE)",
                        "MERCED R",
                        "MARIPOSA",
                        -119.773,
                        37.765,
                        7000,
                        "Yosemite National Park",
                        np.nan,
                    ),
                    (
                        "TUM",
                        "TUOLUMNE MEADOWS",
                        "TUOLUMNE R",
                        "TUOLUMNE",
                        -119.350,
                        37.873,
                        8600,
                        "CA Dept of Water Resources/DFM-Hydro-SMN",
                        np.nan,
                    ),
                    (
                        "SLI",
                        "SLIDE CANYON",
                        "TUOLUMNE R",
                        "TUOLUMNE",
                        -119.43,
                        38.092,
                        9200,
                        "CA Dept of Water Resources/DFM-Hydro-SMN",
                        np.nan,
                    ),
                ],
                columns=[
                    "ID",
                    "Station Name",
                    "River Basin",
                    "County",
                    "Longitude",
                    "Latitude",
                    "ElevationFeet",
                    "Operator",
                    "Map",
                ],
            )
        ]

    def test_class_variables(self):
        assert CDECPointData("no", "no").tzinfo == pytz.timezone("US/Pacific")
        # Base implementation should fail
        with pytest.raises(AttributeError):
            PointData("foo", "bar").tzinfo

    def test_get_metadata(self, tny_station):
        with patch("metloom.pointdata.cdec.requests") as mock_requests:
            mock_requests.get.side_effect = self.tny_side_effect
            metadata = tny_station.metadata
            mock_get = mock_requests.get
            assert mock_get.call_count == 1
            mock_get.assert_called_with(
                "http://cdec.water.ca.gov/cdecstation2/CDecServlet/" "getStationInfo",
                params={"stationID": "TNY"},
            )
        expected = gpd.points_from_xy([-119.0], [42.0], z=[1000.0])[0]
        assert expected == metadata

    def test_get_daily_data(self, tny_station, tny_daily_expected):
        with patch("metloom.pointdata.cdec.requests") as mock_requests:
            mock_requests.get.side_effect = self.tny_side_effect
            response = tny_station.get_daily_data(
                datetime(2021, 5, 16),
                datetime(2021, 5, 18),
                [CdecStationVariables.PRECIPITATIONACCUM],
            )
            mock_get = mock_requests.get
            mock_get.assert_any_call(
                "http://cdec.water.ca.gov/dynamicapp/req/JSONDataServlet",
                params={
                    "Stations": "TNY",
                    "dur_code": "D",
                    "Start": "2021-05-16T00:00:00",
                    "End": "2021-05-18T00:00:00",
                    "SensorNums": "2",
                },
            )
            assert mock_get.call_count == 2
        pd.testing.assert_frame_equal(response, tny_daily_expected)

    def test_points_from_geometry(self, shape_obj, station_search_response):
        expected_url = (
            "https://cdec.water.ca.gov/dynamicapp/staSearch?"
            "sta=&sensor_chk=on&sensor=3"
            "&collect=NONE+SPECIFIED&dur="
            "&active_chk=on&active=Y"
            "&loc_chk=on&lon1=-119.8"
            "&lon2=-119.2&lat1=37.7"
            "&lat2=38.2"
            "&elev1=-5&elev2=99000&nearby=&basin=NONE+SPECIFIED"
            "&hydro=NONE+SPECIFIED&county=NONE+SPECIFIED"
            "&agency_num=160&display=sta"
        )
        with patch("metloom.pointdata.cdec.pd.read_html") as mock_table_read:
            mock_table_read.return_value = station_search_response
            result = CDECPointData.points_from_geometry(
                shape_obj, [CdecStationVariables.SWE]
            )
            mock_table_read.assert_called_with(expected_url)
            assert len(result) == 5
            assert [st.id for st in result] == ["GIN", "DAN", "TNY", "TUM", "SLI"]

    def test_points_from_geometry_fail(self, shape_obj):
        with patch("metloom.pointdata.cdec.pd") as mock_pd:
            mock_pd.read_html.side_effect = side_effect_error
            result = CDECPointData.points_from_geometry(
                shape_obj, [CdecStationVariables.SWE]
            )
            assert result.points == []

    def test_point_collection_to_dataframe(self, shape_obj, station_search_response):
        with patch("metloom.pointdata.cdec.pd.read_html") as mock_table_read:
            mock_table_read.return_value = station_search_response
            result = CDECPointData.points_from_geometry(
                shape_obj, [CdecStationVariables.SWE]
            )
            assert isinstance(result, PointDataCollection)
            points_df = result.to_dataframe()
            for idp, point in enumerate(result):
                point_row = points_df.iloc[idp]
                assert point.name == point_row["name"]
                assert point.id == point_row["id"]
                assert point.metadata == point_row["geometry"]



from datetime import timezone, timedelta, datetime
from unittest.mock import patch, MagicMock
from collections import OrderedDict

import geopandas as gpd
import pandas as pd
import pytest

from metloom.pointdata import SnotelPointData
from metloom.variables import SnotelVariables
from tests.test_point_data import TestPointData


class MockZeepObject:
    def __init__(self, obj):
        """
        Args:
            obj: dictionary of values
        """
        self.mock_dict = OrderedDict(obj)
        self.__values__ = self.mock_dict
        for k, v in obj.items():
            setattr(self, k, v)

    def __dict__(self):
        return self.mock_dict

    def __getitem__(self, item):
        return self.mock_dict[item]


class TestSnotelPointData(TestPointData):

    @pytest.fixture(scope="class")
    def points(self):
        return gpd.points_from_xy([-107.67552], [37.9339], z=[9800.0])[0]

    @staticmethod
    def snotel_meta_sideeffect(*args, **kwargs):
        """
        Mock out the metadata response
        """
        code = kwargs["stationTriplet"]
        available_stations = {
            "538:CO:SNTL": {
                "actonId": "07M27S",
                "beginDate": "1979-10-01 00:00:00",
                "countyName": "Ouray",
                "elevation": 9800.0,
                "endDate": "2100-01-01 00:00:00",
                "fipsCountryCd": "US",
                "fipsCountyCd": "091",
                "fipsStateNumber": "08",
                "huc": "140200060201",
                "hud": "14020006",
                "latitude": 37.9339,
                "longitude": -107.67552,
                "name": "Idarado",
                "shefId": "IDRC2",
                "stationTriplet": "538:CO:SNTL",
                "stationDataTimeZone": -8.0
            },
            "538:CO:SNOW": {
                "actonId": "07M27S",
                "beginDate": "1979-10-01 00:00:00",
                "countyName": "Ouray",
                "elevation": 9800.0,
                "endDate": "2100-01-01 00:00:00",
                "fipsCountryCd": "US",
                "fipsCountyCd": "091",
                "fipsStateNumber": "08",
                "huc": "140200060201",
                "hud": "14020006",
                "latitude": 37.9339,
                "longitude": -107.67552,
                "name": "Idarado",
                "shefId": "IDRC2",
                "stationTriplet": "538:CO:SNOW",
            },
            "FFF:CA:SNOW": {
                "actonId": None,
                "beginDate": "1930-02-01 00:00:00",
                "countyName": "Tuolumne",
                "elevation": 6500.0,
                "endDate": "2100-01-01 00:00:00",
                "fipsCountryCd": "US",
                "fipsCountyCd": "109",
                "fipsStateNumber": "06",
                "huc": "180400090302",
                "latitude": 37.995,
                "longitude": -119.78,
                "name": "Fake1",
                "shefId": None,
                "stationTriplet": "FFF:CA:SNOW",
            },
            "BBB:CA:SNOW": {
                "actonId": None,
                "beginDate": "1948-02-01 00:00:00",
                "countyName": "Tuolumne",
                "elevation": 9300.0,
                "endDate": "2100-01-01 00:00:00",
                "fipsCountryCd": "US",
                "fipsCountyCd": "109",
                "fipsStateNumber": "06",
                "huc": "180400090402",
                "hud": "18040009",
                "latitude": 38.18333,
                "longitude": -119.61667,
                "name": "Fake2",
                "shefId": None,
                "stationTriplet": "BBB:CA:SNOW",
            },
        }
        return MockZeepObject(available_stations[code])

    @staticmethod
    def snotel_data_sideeffect(*args, **kwargs):
        duration = kwargs["duration"]
        if duration == "SEMIMONTHLY":
            return [
                MockZeepObject({
                    'beginDate': '2020-01-20 00:00:00',
                    'collectionDates': ['2020-01-28', '2020-02-27'],
                    'duration': 'SEMIMONTHLY',
                    'endDate': '2020-03-14 00:00:00',
                    'flags': ['V', 'V'], 'stationTriplet': '538:CO:SNTL',
                    'values': [13.19, 13.17]})
            ]
        if duration == "DAILY":
            return [MockZeepObject({
                'beginDate': '2020-03-20 00:00:00',
                'collectionDates': [], 'duration': 'DAILY',
                'endDate': '2020-03-22 00:00:00', 'flags': ['V', 'V', 'V'],
                'stationTriplet': '538:CO:SNTL',
                'values': [13.19, 13.17, 13.14]})]

    @staticmethod
    def snotel_hourly_sideeffect(*args, **kwargs):
        return [
            {
                'beginDate': '2020-01-02 00:00', 'endDate': '2020-01-20 00:00',
                'stationTriplet': '538:CO:SNTL',
                'values': [
                    {
                        'dateTime': '2020-03-20 00:00',
                        'flag': 'V',
                        'value': 13.19
                    }, {
                        'dateTime': '2020-03-20 01:00',
                        'flag': 'V',
                        'value': 13.17
                    }, {
                        'dateTime': '2020-03-20 02:00',
                        'flag': 'V',
                        'value': 13.14
                    }]}]

    @pytest.fixture(scope="class")
    def mock_elements(self):
        return [
            MockZeepObject(
                {
                    'beginDate': '1980-07-23 00:00:00', 'dataPrecision': 1,
                    'duration': 'DAILY', 'elementCd': 'WTEQ',
                    'endDate': '2100-01-01 00:00:00', 'heightDepth': None,
                    'ordinal': 1,
                    'originalUnitCd': 'in', 'stationTriplet': '538:CO:SNTL',
                    'storedUnitCd': 'in'}),
            MockZeepObject(
                {'beginDate': '1980-07-23 00:00:00', 'dataPrecision': 1,
                 'duration': 'SEMIMONTHLY', 'elementCd': 'WTEQ',
                 'endDate': '2100-01-01 00:00:00', 'heightDepth': None,
                 'ordinal': 1,
                 'originalUnitCd': 'in', 'stationTriplet': '538:CO:SNTL',
                 'storedUnitCd': 'in'}),
            MockZeepObject(
                {'beginDate': '1979-10-01 00:00:00', 'dataPrecision': 1,
                 'duration': 'HOURLY', 'elementCd': 'WTEQ',
                 'endDate': '2100-01-01 00:00:00', 'heightDepth': None,
                 'ordinal': 1,
                 'originalUnitCd': 'in', 'stationTriplet': '538:CO:SNTL',
                 'storedUnitCd': 'in'}),
            MockZeepObject(
                {'beginDate': '1980-07-23 00:00:00', 'dataPrecision': 1,
                 'duration': 'MONTHLY', 'elementCd': 'WTEQ',
                 'endDate': '2100-01-01 00:00:00', 'heightDepth': None,
                 'ordinal': 1,
                 'originalUnitCd': 'in', 'stationTriplet': '538:CO:SNTL',
                 'storedUnitCd': 'in'})]

    @pytest.fixture(scope="class")
    def mock_zeep_client(self, mock_elements):
        with patch("metloom.pointdata.snotel_client.zeep.Client") as mock_client:
            mock_service = MagicMock()
            # setup the individual services
            mock_service.getStationMetadata.side_effect = self.snotel_meta_sideeffect
            mock_service.getStationElements.return_value = mock_elements
            mock_service.getStations.return_value = ["FFF:CA:SNOW",
                                                     "BBB:CA:SNOW"]
            mock_service.getData.side_effect = self.snotel_data_sideeffect
            mock_service.getHourlyData.side_effect = self.snotel_hourly_sideeffect
            # assign service to client
            mock_client.return_value.service = mock_service
            yield mock_client

    def test_metadata(self, mock_zeep_client):
        obj = SnotelPointData("538:CO:SNTL", "eh")
        assert (
            obj.metadata
            == gpd.points_from_xy([-107.67552], [37.9339], z=[9800.0])[0]
        )
        assert obj.tzinfo == timezone(timedelta(hours=-8.0))

    @pytest.mark.parametrize(
        "station_id, dts, expected_dts, vals, d1, d2, fn_name",
        [
            (
                "538:CO:SNTL",
                ["2020-03-20 00:00", "2020-03-20 01:00", "2020-03-20 02:00"],
                ["2020-03-20 08:00", "2020-03-20 09:00", "2020-03-20 10:00"],
                [13.19, 13.17, 13.14],
                datetime(2020, 3, 20, 0),
                datetime(2020, 3, 20, 2),
                "get_hourly_data",
            ),
            (
                "538:CO:SNTL",
                ["2020-03-20", "2020-03-21", "2020-03-22"],
                ["2020-03-20 08:00", "2020-03-21 08:00", "2020-03-22 08:00"],
                [13.19, 13.17, 13.14],
                datetime(2020, 3, 20),
                datetime(2020, 3, 22),
                "get_daily_data",
            ),
            (
                "538:CO:SNOW",
                ["2020-01-28", "2020-02-27"],
                ["2020-01-28 00:00", "2020-02-27 00:00"],
                [13.19, 13.17],
                datetime(2020, 1, 20),
                datetime(2020, 3, 15),
                "get_snow_course_data",
            ),
        ],
    )
    def test_get_data_methods(self, station_id, dts, expected_dts, vals, d1,
                              d2,
                              fn_name, points, mock_zeep_client):
        station = SnotelPointData(station_id, "TestSite")
        vrs = [SnotelVariables.SWE]
        fn = getattr(station, fn_name)
        result = fn(d1, d2, vrs)
        expected = self.expected_response(
            expected_dts, vals, SnotelVariables.SWE.name, "in", station, points
        )
        pd.testing.assert_frame_equal(result, expected)

    def test_points_from_geometry(self, shape_obj, mock_zeep_client):
        result = SnotelPointData.points_from_geometry(
            shape_obj, [SnotelVariables.SWE], snow_courses=True
        )
        ids = [point.id for point in result]
        names = [point.name for point in result]
        assert len(names) == 2
        assert set(ids) == {"FFF:CA:SNOW", "BBB:CA:SNOW"}
        assert set(names) == {"Fake1", "Fake2"}

    def test_points_from_geometry_fail(self, shape_obj, mock_zeep_client):
        mock_zeep_client.return_value.service.getStations.return_value = []
        result = SnotelPointData.points_from_geometry(
            shape_obj, [SnotelVariables.SWE], snow_courses=True
        )
        assert len(result) == 0



from datetime import datetime, timezone, timedelta
from typing import List, Dict
import logging
import geopandas as gpd
import pandas as pd
from functools import reduce


from .base import PointData
from ..variables import SnotelVariables, SensorDescription
from ..dataframe_utils import join_df, append_df

from .snotel_client import (
    DailySnotelDataClient, MetaDataSnotelClient, HourlySnotelDataClient,
    SemiMonthlySnotelClient, PointSearchSnotelClient, SeriesSnotelClient,
    ElementSnotelClient
)

LOG = logging.getLogger("metloom.pointdata.snotel")


class SnotelPointData(PointData):
    """
    Implement PointData methods for SNOTEL data source
    API documentation here:
    https://www.nrcs.usda.gov/wps/portal/wcc/home/dataAccessHelp/webService/webServiceReference/
    Website has variable and network codes
    Possible testing resource:
    https://wcc.sc.egov.usda.gov/awdbWebService/webservice/testwebservice.jsf?webserviceName=/awdbWebService
    """

    ALLOWED_VARIABLES = SnotelVariables

    def __init__(self, station_id, name, metadata=None):
        """
        See docstring for PointData.__init__
        """
        super(SnotelPointData, self).__init__(
            station_id, name, metadata=metadata
        )
        self._raw_metadata = None
        self._raw_elements = None
        self._tzinfo = None

    def _snotel_response_to_df(self, result_map: Dict[SensorDescription, List[dict]],
                               duration: str):
        """
        Convert the response from climata.snotel classes into
        Args:
           result_map: map of the sensors to the list of API results
        """
        df = None
        # TODO: possible DRY opportunity here too
        final_columns = ["geometry", "site", "measurementDate"]
        for variable, data in result_map.items():
            transformed = [
                {
                    "datetime": row["datetime"],
                    "site": self.id,
                    "measurementDate": row["datetime"],
                    variable.name: row["value"],
                    f"{variable.name}_units": self._get_units(variable, duration),
                }
                for row in data
            ]
            final_columns += [variable.name, f"{variable.name}_units"]
            sensor_df = gpd.GeoDataFrame.from_dict(
                transformed, geometry=[self.metadata] * len(transformed)
            )
            # TODO: possibly an opportunity for DRY here (see CDEC)
            sensor_df["datetime"] = pd.to_datetime(sensor_df["datetime"])
            sensor_df["measurementDate"] = pd.to_datetime(
                sensor_df["measurementDate"]
            )
            sensor_df["datetime"] = sensor_df["datetime"].apply(
                self._handle_df_tz
            )
            sensor_df["measurementDate"] = sensor_df["measurementDate"].apply(
                self._handle_df_tz
            )
            # set index so joining works
            sensor_df.set_index("datetime", inplace=True)
            sensor_df = sensor_df.filter(final_columns)
            df = join_df(df, sensor_df)

        if df is not None and len(df.index) > 0:
            df.reset_index(inplace=True)
            df.set_index(keys=["datetime", "site"], inplace=True)
            df.index.set_names(["datetime", "site"], inplace=True)
        self.validate_sensor_df(df)
        return df

    def _fetch_data_for_variables(self, client: SeriesSnotelClient,
                                  variables: List[SensorDescription],
                                  duration: str):
        result_map = {}
        for variable in variables:
            data = client.get_data(element_cd=variable.code)
            if len(data) > 0:
                result_map[variable] = data
            else:
                LOG.warning(f"No {variable.name} found for {self.name}")
        return self._snotel_response_to_df(result_map, duration)

    def get_daily_data(
        self,
        start_date: datetime,
        end_date: datetime,
        variables: List[SensorDescription],
    ):
        """
        See docstring for PointData.get_daily_data
        """
        client = DailySnotelDataClient(
            station_triplet=self.id,
            begin_date=start_date,
            end_date=end_date,
        )
        return self._fetch_data_for_variables(client, variables, client.DURATION)

    def get_hourly_data(
        self,
        start_date: datetime,
        end_date: datetime,
        variables: List[SensorDescription],
    ):
        """
        See docstring for PointData.get_hourly_data
        """
        client = HourlySnotelDataClient(
            station_triplet=self.id,
            begin_date=start_date,
            end_date=end_date,
        )
        return self._fetch_data_for_variables(client, variables, "HOURLY")

    def get_snow_course_data(
        self,
        start_date: datetime,
        end_date: datetime,
        variables: List[SensorDescription],
    ):
        """
        See docstring for PointData.get_snow_course_data
        """
        client = SemiMonthlySnotelClient(
            station_triplet=self.id,
            begin_date=start_date,
            end_date=end_date,
        )
        return self._fetch_data_for_variables(client, variables, client.DURATION)

    def _get_all_metadata(self):
        """
        Set _raw_metadata once using Snotel API
        """
        if self._raw_metadata is None:
            client = MetaDataSnotelClient(station_triplet=self.id)
            self._raw_metadata = client.get_data()
        return self._raw_metadata

    def _get_all_elements(self):
        """
        Set _raw_metadata once using Snotel API
        """
        if self._raw_elements is None:
            client = ElementSnotelClient(station_triplet=self.id)
            self._raw_elements = client.get_data()
        return self._raw_elements

    def _get_units(self, variable: SensorDescription, duration: str):
        units = None
        for meta in self._get_all_elements():
            if meta["elementCd"] == variable.code and meta["duration"] == duration:
                units = meta["storedUnitCd"]
                break
        if units is None:
            raise ValueError(f"Could not find units for {variable}")
        return units

    def _get_metadata(self):
        """
        See docstring for PointData._get_metadata
        """
        all_metadata = self._get_all_metadata()
        if isinstance(all_metadata, list):
            data = all_metadata[0]
        else:
            data = all_metadata
        return gpd.points_from_xy(
            [data["longitude"]], [data["latitude"]], z=[data["elevation"]]
        )[0]

    def _get_tzinfo(self):
        """
        Return timezone info that pandas can use from the raw_metadata
        """
        metadata = self._get_all_metadata()
        # Snow courses might not have a timezone attached
        tz_hours = metadata.get("stationDataTimeZone")
        if tz_hours is None:
            LOG.error(f"Could not find timezone info for {self.id} ({self.name})")
            tz_hours = 0
        else:
            tz_hours = float(tz_hours)
        return timezone(timedelta(hours=tz_hours))

    @property
    def tzinfo(self):
        if self._tzinfo is None:
            self._tzinfo = self._get_tzinfo()
        return self._tzinfo

    @classmethod
    def points_from_geometry(
        cls,
        geometry: gpd.GeoDataFrame,
        variables: List[SensorDescription],
        snow_courses=False,
    ):
        """
        See docstring for PointData.points_from_geometry
        """
        projected_geom = geometry.to_crs(4326)
        bounds = projected_geom.bounds.iloc[0]
        # TODO: network may need to change to get streamflow
        network = "SNOW" if snow_courses else "SNTL"
        point_codes = []
        for variable in variables:
            # this search is default AND on all parameters
            # so search for each variable seperately
            response = PointSearchSnotelClient(
                max_latitude=bounds["maxy"],
                min_latitude=bounds["miny"],
                max_longitude=bounds["maxx"],
                min_longitude=bounds["minx"],
                network_cds=network,
                element_cds=variable.code
                # ordinals=  # TODO: what are ordinals?
            ).get_data()
            if len(response) > 0:
                point_codes += response

        # no duplicate codes
        point_codes = list(set(point_codes))
        dfs = [
            pd.DataFrame.from_records(
                [MetaDataSnotelClient(station_triplet=code).get_data()]
            ).set_index("stationTriplet") for code in point_codes
        ]

        if len(dfs) > 0:
            df = reduce(lambda a, b: append_df(a, b), dfs)
        else:
            return cls.ITERATOR_CLASS([])

        df.reset_index(inplace=True)
        gdf = gpd.GeoDataFrame(
            df,
            geometry=gpd.points_from_xy(
                df["longitude"], df["latitude"], z=df["elevation"]
            ),
        )
        filtered_gdf = gdf[gdf.within(projected_geom.iloc[0]["geometry"])]
        points = [
            cls(row[0], row[1], metadata=row[2])
            for row in zip(
                filtered_gdf["stationTriplet"],
                filtered_gdf["name"],
                filtered_gdf["geometry"],
            )
        ]
        return cls.ITERATOR_CLASS(points)
